@@ -6,7 +6,7 @@
  *
  * -----------------------------------
  *
- * Copyright (C) 2002 - 2011 Phil Dibowitz
+ * Copyright (C) 2002 - present Phil Dibowitz
  *
  * This software is provided 'as-is', without any express or
  * implied warranty. In no event will the authors be held
@@ -63,7 +63,7 @@ extern "C" {
 #include <unistd.h>
 using namespace std;
 
-#define VERSION "2.2.3"
+#define VERSION "2.2.5"
 /*
  * MAXCONS is set to 16k, the default number of states in iptables. Generally
  * speaking the ncurses pad is this many lines long, but since ncurses
@@ -336,14 +336,51 @@ void resolve_port(const unsigned int &port, string &name, const string &proto)
   }
 }
 
-void resolve_names(table_t &entry, max_t &max)
+/*
+ * If lookup mode is on, we lookup the names and put them in the structure.
+ *
+ * If lookup mode is not on, we generate strings of the addresses and put
+ * those in the structure.
+ *
+ * Finally, we update the max_t structure.
+ *
+ * NOTE: We stringify addresses largely because in the IPv6 case we need
+ * to treat them like truncate-able strings.
+ */
+void stringify_entry(table_t &entry, max_t &max, const flags_t &flags)
 {
   unsigned int size = 0;
+  ostringstream buffer;
+  char tmp[NAMELEN];
 
-  resolve_host(entry.family, entry.src, entry.sname);
-  resolve_host(entry.family, entry.dst, entry.dname);
-  resolve_port(entry.srcpt, entry.spname, entry.proto);
-  resolve_port(entry.dstpt, entry.dpname, entry.proto);
+  bool have_port = entry.proto == "tcp" || entry.proto == "udp";
+  if (!have_port) {
+    entry.spname = entry.dpname = "";
+  }
+
+  if (flags.lookup) {
+    resolve_host(entry.family, entry.src, entry.sname);
+    resolve_host(entry.family, entry.dst, entry.dname);
+    if (have_port) {
+      resolve_port(entry.srcpt, entry.spname, entry.proto);
+      resolve_port(entry.dstpt, entry.dpname, entry.proto);
+    }
+  } else {
+    buffer << inet_ntop(entry.family, (void*)&entry.src, tmp, NAMELEN-1);
+    entry.sname = buffer.str();
+    buffer.str("");
+    buffer << inet_ntop(entry.family, (void*)&entry.dst, tmp, NAMELEN-1);
+    entry.dname = buffer.str();
+    buffer.str("");
+    if (have_port) {
+      buffer << entry.srcpt;
+      entry.spname = buffer.str();
+      buffer.str("");
+      buffer << entry.dstpt;
+      entry.dpname = buffer.str();
+      buffer.str("");
+    }
+  }
 
   size = entry.sname.size() + entry.spname.size() + 1;
   if (size > max.src)
@@ -352,29 +389,6 @@ void resolve_names(table_t &entry, max_t &max)
   size = entry.dname.size() + entry.dpname.size() + 1;
   if (size > max.dst)
     max.dst = size;
-
-}
-
-
-/*
- * Based on the format pre-chosen, truncate src/dst as needed.
- */
-void truncate(table_t &table, const max_t &max, const flags_t &flags)
-{
-  int length;
-  if (table.sname.size() + table.spname.size() + 1 > max.src) {
-    length = max.src - 1 - table.spname.size();
-    table.sname = table.sname.substr(0, length);
-    if (flags.tag_truncate)
-      table.sname[table.sname.size()-1] = '+';
-  }
-
-  if (table.dname.size() + table.dpname.size() + 1 > max.dst) {
-    length = max.dst - 1 - table.dpname.size();
-    table.dname = table.dname.substr(0, length);
-    if (flags.tag_truncate)
-      table.dname[0] = '+';
-  }
 }
 
 
@@ -791,7 +805,8 @@ void initialize_maxes(max_t &max, flags_t &flags)
     max.src = max.dst = 21;
   }
   /*
-   * The proto header is 5, so we can't drop below 6.
+   * The proto starts at 3, since tcp/udp are the most common, but will
+   * grow if we see bigger proto strings such as "ICMP".
    */
   max.proto = 3;
   /*
@@ -895,13 +910,19 @@ void delete_state(WINDOW *&win, const table_t &entry, const flags_t &flags)
     nfct_set_attr(ct, ATTR_ORIG_IPV6_DST, (void *)&entry.dst.s6_addr);
   }
 
-  nfct_set_attr_u8(ct, ATTR_ORIG_L4PROTO,
-                   getprotobyname(entry.proto.c_str())->p_proto);
+  // undo our space optimization so the kernel can find the state.
+  protoent *pn;
+  if (entry.proto == "icmp6")
+    pn = getprotobyname("ipv6-icmp");
+  else
+    pn = getprotobyname(entry.proto.c_str());
+
+  nfct_set_attr_u8(ct, ATTR_ORIG_L4PROTO, pn->p_proto);
 
   if (entry.proto == "tcp" || entry.proto == "udp") {
     nfct_set_attr_u16(ct, ATTR_ORIG_PORT_SRC, htons(entry.srcpt));
     nfct_set_attr_u16(ct, ATTR_ORIG_PORT_DST, htons(entry.dstpt));
-  } else if (entry.proto == "icmp") {
+  } else if (entry.proto == "icmp" || entry.proto == "icmp6") {
     string type, code, id, tmp;
     split('/', entry.state, type, tmp);
     split(' ', tmp, code, tmp);
@@ -955,7 +976,7 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
   struct protoent* pe = NULL;
   int seconds, minutes, hours;
   char ttlc[11];
-  ostringstream typecode;
+  ostringstream buffer;
 
   /*
    * Clear the entry
@@ -973,11 +994,20 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
    * same regardless of protocol
    */
 
-  pe = getprotobynumber(nfct_get_attr_u8(ct, ATTR_ORIG_L4PROTO));
+  short int pr = nfct_get_attr_u8(ct, ATTR_ORIG_L4PROTO);
+  pe = getprotobynumber(pr);
   if (pe == NULL) {
-    entry.proto = "unknown";
+    buffer << pr;
+    entry.proto = buffer.str();
+    buffer.str("");
   } else {
     entry.proto = pe->p_name;
+    /* 
+     * if proto is "ipv6-icmp" we can just say "icmp6" to save space...
+     * it's more common/standard anyway
+     */
+    if (entry.proto == "ipv6-icmp")
+      entry.proto = "icmp6";
   }
 
   // ttl
@@ -1035,13 +1065,15 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
   } else if (entry.proto == "udp") {
     entry.state = "";
     counts->udp++;
-  } else if (entry.proto == "icmp") {
-    typecode.str("");
-    typecode << (int)nfct_get_attr_u8(ct, ATTR_ICMP_TYPE) << "/"
+  } else if (entry.proto == "icmp" || entry.proto == "icmp6") {
+    buffer.str("");
+    buffer << (int)nfct_get_attr_u8(ct, ATTR_ICMP_TYPE) << "/"
         << (int)nfct_get_attr_u8(ct, ATTR_ICMP_CODE) << " ("
         << nfct_get_attr_u16(ct, ATTR_ICMP_ID) << ")";
-    entry.state = typecode.str();
+    entry.state = buffer.str();
     counts->icmp++;
+    if (entry.state.size() > max->state)
+      max->state = entry.state.size();
   } else {
     counts->other++;
   }
@@ -1093,12 +1125,12 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
   }
 
   /*
-   * RESOLVE AND TRUNCATE
+   * RESOLVE
    */
 
-  // Resolve Names if we need to
-  if (flags->lookup)
-    resolve_names(entry,*max);
+  // Resolve names - if necessary - or generate strings of address,
+  // and calculate max sizes
+  stringify_entry(entry, *max, *flags);
 
   /*
    * Add this to the array
@@ -1373,6 +1405,69 @@ void print_headers(const flags_t &flags, const string &format,
 
 }
 
+
+void truncate(string &string, int length, bool mark, char direction)
+{
+  int s = (direction == 'f') ? string.size() - length : 0;
+
+  string = string.substr(s, length);
+  if (mark) {
+    int m = (direction == 'f') ? 0 : string.size() - 1;
+    string[m] = '+';
+  }
+}
+
+/*
+ * Based on the format pre-chosen, truncate src/dst as needed, and then
+ * generate the host:port strings and drop them off in the src/dst string
+ * objects passed in.
+ */
+void format_src_dst(table_t &table, string &src, string &dst,
+                      const flags_t &flags, const max_t &max)
+{
+  ostringstream buffer;
+  bool have_port = table.proto == "tcp" || table.proto == "udp";
+  char direction;
+  unsigned int length;
+
+  // What length would we currently use?
+  length = table.sname.size();
+  if (have_port)
+    length += table.spname.size() + 1;
+
+  // If it's too long, figure out how room we have and truncate it
+  if (length > max.src) {
+    length = max.src;
+    if (have_port)
+      length -= 1 + table.spname.size();
+    direction = (flags.lookup) ? 'e' : 'f';
+    truncate(table.sname, length, flags.tag_truncate, direction);
+  }
+
+  // ... and repeat
+  length = table.dname.size();
+  if (have_port)
+    length += table.dpname.size() + 1;
+  if (length > max.dst) {
+    length = max.dst;
+    if (have_port)
+      length -= 1 + table.dpname.size();
+    direction = (flags.lookup) ? 'f' : 'e';
+    truncate(table.dname, length, flags.tag_truncate, direction);
+  }
+
+  buffer << table.sname;
+  if (have_port)
+    buffer << ":" << table.spname;
+  src = buffer.str();
+  buffer.str("");
+  buffer << table.dname;
+  if (have_port)
+    buffer << ":" << table.dpname;
+  dst = buffer.str();
+  buffer.str("");
+}
+
 /*
  * An abstraction of priting a line for both single/curses modes
  */
@@ -1382,37 +1477,9 @@ void printline(table_t &table, const flags_t &flags, const string &format,
   ostringstream buffer;
   buffer.str("");
   string src, dst, b, p;
-  char tmp[NAMELEN];
   
-  if (flags.lookup)
-    truncate(table, max, flags);
-
-  if (table.proto == "tcp" || table.proto == "udp") {
-    if (flags.lookup && (table.sname != ""))
-      buffer << table.sname << ":" << table.spname;
-    else
-      buffer << inet_ntop(table.family, (void*)&table.src, tmp, NAMELEN-1)
-          << ":" << table.srcpt;
-    src = buffer.str();
-    buffer.str("");
-    if (flags.lookup && (table.dname != "")) {
-      buffer << table.dname << ":" << table.dpname;
-    } else {
-      buffer << inet_ntop(table.family, (void*)&table.dst, tmp, NAMELEN-1)
-        << ":" << table.dstpt;
-    }
-    dst = buffer.str();
-    buffer.str("");
-  } else {
-    if (flags.lookup && (table.sname != ""))
-      src = table.sname;
-    else
-      src = inet_ntop(table.family, (void*)&table.src, tmp, NAMELEN-1);
-    if (flags.lookup && (table.dname != ""))
-      dst = table.dname;
-    else
-      dst = inet_ntop(table.family, (void*)&table.dst, tmp, NAMELEN-1);
-  }
+  // Generate strings for src/dest, truncating and marking as necessary
+  format_src_dst(table, src, dst, flags, max);
 
   if (flags.counters) {
     buffer << table.bytes;
@@ -1437,7 +1504,7 @@ void printline(table_t &table, const flags_t &flags, const string &format,
         color = 1;
       else if (table.proto == "udp")
         color = 2;
-      else if (table.proto == "icmp")
+      else if (table.proto == "icmp" || table.proto == "icmp6")
         color = 3;
       if (curr)
         color += 4;
@@ -2553,8 +2620,7 @@ int main(int argc, char *argv[])
         wclrtoeol(mainwin);
         break;
       case 'x':
-        delete_state(mainwin, stable[curr_state],
-            flags);
+        delete_state(mainwin, stable[curr_state], flags);
         break;
       /*
        * Window navigation
